*:*****************************************************************************
*:
*: Procedure file: RPROFILE.PRG
*:
*:         System: RProfile : MS-Windows like .INI file functions
*:         Author: Rolf van Gelder
*:                 Copyright (c) 1994-95, Rolf van Gelder
*:                    RvG Clipper Collection (tm)
*:                    All rights reserved
*:        
*:  Last modified: 21/07/95
*:
*: Donated to the Public Domain by Rolf van Gelder
*:
*: To create a stand-alone test program : CLIPPER RPROFILE /n/dTEST
*:    
*:*****************************************************************************

#include "FileIO.ch"

#define	CR_LF	Chr(13) + Chr(10)

#translate SetDefault (<var> , <default>) => ;
   <var> := IF ( <var> = nil , <default> , <var> )

*-- File cache buffer
STATIC	cCache

*-- Name of the .INI file that is in the cache buffer
STATIC	cIniFile

*-- Read complete file into cache buffer
STATIC	bRead := ;
   { |hnd| cCache := Space ( FSeek ( hnd, 0, FS_END ) ), ;
   FSeek ( hnd, 0, FS_SET ), ;
   FRead ( hnd, @cCache, Len ( cCache ) ) }

*-- Write section to file
STATIC	bPutSection := ;
   { |hnd,sect| IF ( !Empty ( sect ), ;
   FWrite ( hnd, '[' + sect + ']' + CR_LF ), ;
   nil) }

*-- Write entry to file
STATIC	bPutEntry := ;
   { |hnd,entry,val| IF ( !Empty ( entry ) .and. !Empty ( val ), ;
   FWrite ( hnd, entry + '=' + val + CR_LF ), ;
   nil) }

*-- Rewrite complete file from buffer
STATIC	bReWrite := ;
   { |hnd,fnm| IF ( hnd!=NIL, FClose(hnd), NIL ), ;
   hnd := FCreate ( fnm, FC_NORMAL ), ;
   FWrite ( hnd, cCache ), ;
   FClose ( hnd ) }

*-- Look for the specified section in buffer
STATIC	bSeekSection := ;
   { |sect,pos| (pos:=At ('['+Upper (sect)+']',Upper (cCache) ) ) > 0 }


*!*****************************************************************************
*!
*!       Function: TEST()
*!
*!*****************************************************************************

#IFDEF TEST

	FUNCTION Test ()

	LOCAL	aEntries := {}

	*-- Add items to the file PROFILE.INI
	R_IniWrite ( 'Date & Time', 'Date', DToC ( Date () ), 'PROFILE.INI' )
	R_IniWrite ( 'Date & Time', 'Time', Time (), 'PROFILE.INI' )

	*-- Delete [Date & Time] - all entries
	* R_IniWrite ( 'Date & Time', NIL, NIL, 'PROFILE.INI' )

	*-- Delete [Date & Time].Time entry
	* R_IniWrite ( 'Date & Time', 'Time', NIL, 'PROFILE.INI' )

	*-- Get all entries from the [Date & Time] section
	* aEntries := R_IniRead ( 'Date & Time', nil, '00:00', 'PROFILE.INI' )
	*
	* IF aEntries = NIL
	*    Alert ( 'No entries found.' )
	* ELSE
	*    *-------------------------     Entry        Value
	*    AEval ( aEntries, { |x| QOut ( x[1] + '=' + x[2] ) } )
	* ENDIF

	*-- Get the Day entry from the [Date & Time] section
	*-- create the entry if not present and set the value to the
	*-- default value
	R_IniRead ( 'Date & Time', 'Day', CDow ( Date () ), 'PROFILE.INI', .T. )

	RETURN nil

#ENDIF


*!*****************************************************************************
*!
*!       Function: R_INIREAD()
*!
*!*****************************************************************************
FUNCTION R_IniRead ( cSection, cEntry, cDefault, cFName, lAppend )

*--------------------------------------------------------------------------
* This function searches the file for an entry that matches the name
* specified by the cEntry parameter under the section heading specified
* by the cSection parameter.
* If the entry is found, it corresponding string is returned.
* If the entry does not exist, the default string specified by the cDefault
* parameter is copied. If, in that case, the lAppend flag is .T. the default
* entry will be added to the file.
*
* cSection : String that specifies the section containing the entry.
* cEntry   : String containing the entry whose associated string is to be
*            retrieved. If this value is NIL, all entries in the section
*            specified by the cSection parameter are copied to an array.
* cDefault : String that specifies the default value for the given entry
*            if the entry cannot be found in the initialization file.
* cFName   : String that names the initialization file
* lAppend  : Boolean that specifies if the entry (with the associated
*            default value) should be added to the INI file if the entry
*            cannot be found.
*            If the section does not exist it will be created.
*--------------------------------------------------------------------------

LOCAL	nHandle
LOCAL	cString
LOCAL	nPos
LOCAL	nEnd
LOCAL	aEntries := {}

IF cSection = NIL .or.   Valtype ( cSection ) != 'C'
   *-- Required parameter
   RETURN nil
ENDIF

IF cEntry != NIL   .and. Valtype ( cEntry ) != 'C'
   *-- Invalid type
   RETURN nil
ENDIF

IF cDefault != NIL .and. Valtype ( cDefault ) != 'C'
   *-- Invalid type
   RETURN nil
ENDIF

IF cFName = NIL .or. Valtype ( cFName ) != 'C'
   *-- Required parameter
   RETURN nil
ENDIF

SetDefault ( lAppend, .F. )

IF Valtype ( lAppend ) != 'L'
   *-- Invalid type
   RETURN nil
ENDIF

*-- Extension omitted : Add default extension
IF At ( '.', cFName ) = 0
   cFName -= '.INI'
ENDIF

IF cIniFile = NIL
   *-- First time ... (buffer not in cache)
   cIniFile := ''
ENDIF

*-- Check the filename
IF cIniFile != cFName
   *-- Other .INI file name : file not in cache !
   
   IF ( nHandle := FOpen ( cFName, FO_READ + FO_SHARED ) ) < 5
      *-- Error opening .INI file
      RETURN nil
   ENDIF
   
   *-- INI file opened ...
   
   *-- Read complete file into cCache
   Eval ( bRead, nHandle )
   
   *-- File not needed anymore ...
   FClose ( nHandle )
   
   *-- File in cache ...
   cIniFile := cFName
   
ENDIF

IF Eval ( bSeekSection, cSection, @nPos )
   *-- Section FOUND, nPos points to the start of the section
   
   IF cEntry = NIL
      *-- RETURN ALL ENTRIES IN SECTION !
      
      *-- nPos points to start of section
      
      *-- Skip [section] + CR_LF
      nPos = nPos + 4 + Len ( cSection )
      
      *-- nPos points to start of first entry
      DO WHILE nPos <= Len ( cCache ) .and. SubStr ( cCache, nPos, 1 ) != '['
         
         nEnd    := I_At ( '=', cCache, nPos )
         cEntry  := SubStr ( cCache, nPos, nEnd - nPos )
         nPos    := nEnd + 1
         nEnd    := I_At ( CR_LF, cCache, nPos )
         cString := SubStr ( cCache, nPos, nEnd - nPos )
         AAdd ( aEntries, { cEntry, cString } )
         nPos    := nEnd + 2
         
         DO WHILE SubStr ( cCache, nPos, 2 ) = CR_LF
            *-- Skip CR_LF's, if any
            nPos += 2
         ENDDO
         
      ENDDO
      
      RETURN aEntries
      
   ELSE
      *-- Locate specified entry
      nPos := I_At ( Upper ( cEntry ) + '=', Upper ( cCache ), nPos )
      
      IF nPos > 0
         *-- Entry found, nPos points to start of entry
         
         *-- Skip 'entry=' part
         nPos += Len ( cEntry )
         
         *-- Return value
         RETURN SubStr ( cCache, nPos+1, ;
            I_At ( CR_LF, cCache, nPos+1 ) - nPos - 1 )
         
      ENDIF
      
   ENDIF
   
ELSE
   *-- Section not found
   
   IF cEntry = NIL
      *-- Request to return all entries in section ...
      RETURN nil
   ENDIF
   
ENDIF

IF lAppend
   *-- CREATE A NEW cDEFAULT ENTRY, IF THE SPECIFIED ENTRY NOT EXISTS
   
   R_IniWrite ( cSection, cEntry, cDefault, cIniFile )
   
ENDIF

*-- Return default value
RETURN cDefault


*!*****************************************************************************
*!
*!       Function: R_INIWRITE()
*!
*!*****************************************************************************
FUNCTION R_IniWrite ( cSection, cEntry, cString, cFName )

*--------------------------------------------------------------------------
* cSection : String that specifies the section to which the string will
*            will be copied. If the section does not exist, it is created.
*            cSection is case-independent
* cEntry   : String containing the entry to be associated with the string.
*            If the entry does not exist in the specified section, it is
*            created. If the parameter is NIL, the entire section,
*            including all entries within the section, is deleted.
* cString  : String to be written to the file. If this parameter is NIL,
*            the entry specified by the cEntry parameter is deleted.
* cFName   : String that names the initialization file
*--------------------------------------------------------------------------

LOCAL	nHandle
LOCAL	nBytes
LOCAL	nPos
LOCAL	nEntry

IF cFName = NIL .or. Valtype ( cFName ) != 'C'
   *-- Required parameter !
   RETURN .f.
ENDIF

IF cSection = NIL .or. Valtype ( cSection ) != 'C'
   *-- Required paramter !
   RETURN .f.
ENDIF

*-- Append default extension (if no extension present)
IF At ( '.', cFName ) = 0
   cFName -= '.INI'
ENDIF

IF cIniFile = NIL
   *-- First time ...
   cIniFile := ''
ENDIF

*-- Check the filename
IF cIniFile != cFName
   *-- If file name is the SAME : file is still in Cache buffer (cCache) ...
   
   *-- Other .INI file or the first time
   IF ( nHandle := FOpen ( cFName, FO_READWRITE + FO_SHARED ) ) < 5
      *-- Error opening .INI file
      
      IF ( nHandle := FCreate ( cFName, FC_NORMAL ) ) < 5
         *-- Error creating .INI file
         RETURN .f.
         
      ELSE
         *-- .INI file created : Write Section and Entry
         Eval ( bPutSection, nHandle, cSection )
         
         Eval ( bPutEntry, nHandle, cEntry, cString )
         
         *-- ReRead file to adjust cCache cache
         Eval ( bRead, nHandle )
         
         FClose ( nHandle )
         
         *-- Buffer in cache ...
         cIniFile := cFName
         
         RETURN .t.
         
      ENDIF
      
   ENDIF
   
   *-- Read complete file into cCache
   Eval ( bRead, nHandle )
   
   *-- Buffer in cache ...
   cIniFile := cFName
   
ENDIF

IF !Eval ( bSeekSection, cSection, @nPos )
   *-- Section NOT present : append SECTION & ENTRY !
   
   IF nHandle = NIL
      *-- File not presently open
      nHandle := FOpen ( cFName, FO_READWRITE + FO_SHARED )
   ENDIF
   
   *-- Pointer to end-of-file
   FSeek  ( nHandle, 0, FS_END )
   
   *-- APPEND NEW SECTION (separated with an empty line)
   FWrite ( nHandle, CR_LF, 2 )
   Eval ( bPutSection, nHandle, cSection )
   
   Eval ( bPutEntry, nHandle, cEntry, cString )
   
   *-- ReRead file to adjust cCache ....
   Eval ( bRead, nHandle )
   
ELSE
   *-- SECTION ALREADY PRESENT
   
   *-- nPos points to the start of the section
   
   IF cEntry = NIL
      
      *-- DELETE COMPLETE SECTION !
      
      *-- nPos points to start of section
      IF ( nEntry := I_At ( '[', cCache, nPos + 1 ) ) = 0
         *-- No next section : delete to end-of-file
         
         nEntry := Len ( cCache ) + 1
         
      ENDIF
      
      *-- Delete bytes from string
      cCache := ;
         Stuff ( cCache, nPos, nEntry - nPos, '' )
      
      Eval ( bReWrite, nHandle, cFName )
      
      RETURN .t.
      
   ENDIF
   
   *-- Skip section + CR_LF
   nPos = nPos + 4 + Len ( cSection )
   
   IF ( nEntry := I_At ( Upper ( cEntry ) + '=', ;
         Upper ( cCache ), nPos ) ) = 0
      
      *-- ENTRY NOT FOUND : APPEND ENTRY
      IF cString != NIL
         *-- Locate start of next SECTION
         IF nHandle = NIL
            nHandle := FOpen ( cFName, FO_READWRITE + FO_SHARED )
         ENDIF
         
         IF ( nEntry := I_At ( '[', cCache, nPos ) ) = 0
            
            *-- Last section : append to end of file
            FSEEK ( nHandle, 0, FS_END )
            
            Eval ( bPutEntry, nHandle, cEntry, cString )
            
            *-- ReRead file to adjust cCache
            Eval ( bRead, nHandle )
            
         ELSE
            *-- Next section present at : nEntry - 2
            
            *-- INSERT ENTRY AT END OF SECTION
            DO WHILE SubStr ( cCache, nEntry-2, 2 ) = CR_LF
               *-- Skip CR_LF's, if any ...
               nEntry -= 2
            ENDDO
            
            *-- Keep 1 CR_LF string ...
            nEntry += 2
            
            cCache := ;
               Stuff ( cCache, nEntry, 0, cEntry + '=' + cString + CR_LF )
            
            Eval ( bReWrite, nHandle, cFName )
            
            RETURN .t.
            
         ENDIF
         
      ENDIF
      
   ELSE
      *-- ENTRY FOUND : REPLACE VALUE
      
      IF cString = NIL
         *-- DELETE ENTRY !
         
         *-- nEntry points to first pos of entry name
         nPos := I_At ( CR_LF, cCache, nEntry ) + 2
         
         *-- Delete bytes from string
         cCache := ;
            Stuff ( cCache, nEntry, nPos - nEntry, '' )
         
      ELSE
         *-- REPLACE VALUE
         
         *-- nEntry points to first pos of entry name
         
         nEntry  := nEntry + Len ( cEntry ) + 1
         cCache := Stuff ( cCache, nEntry, ;
            At ( CR_LF, SubStr ( cCache, nEntry ) ) - 1, cString )
         
      ENDIF
      
      Eval ( bReWrite, nHandle, cFName )
      
      RETURN .t.
      
   ENDIF
   
ENDIF

FClose ( nHandle )

RETURN .t.


*!*****************************************************************************
*!
*!       Function: I_AT()
*!
*!*****************************************************************************
STATIC FUNCTION I_At ( cSearch, cString, nStart )

*--------------------------------------------------------------------------
* AT () function with a start position to search from
*--------------------------------------------------------------------------

LOCAL	nPos := At ( cSearch, SubStr ( cString, nStart ) )

RETURN IF ( nPos > 0, nPos + nStart - 1, 0 )

*******************************************************************************
* ___________  ________________________________________________________________
*  \_____    \  * * * *   R v G   C l i p p e r   C o l l e c t i o n   * * * *
*   |       _/_  _ ________                    ______               (c) 1993-95
*   \    |   \\ \//  _____/                    \____ \          Rolf van Gelder
*   /____|_  / \//   \  ___                    )  |_> >   Binnenwiertzstraat 27
* _________\/__  \    \_\  \     _THE_         |   __/ost    5615 HG  EINDHOVEN
* \            \__\______  /   ShareWare    ___|__|             The Netherlands
*  \_  ______   \_   ___ \/    Libraries    \____ \ 
*  /   \     \  /    \  \/   for CA-CLIPPER )  | > >              +31-40-438852
* /     \     \/\     \____                 |   __/hone         (Eve & Weekend)
* \      \_______\______  /               __|__|     
*  \             /      \/               / __ \     Internet: rcrolf@urc.tue.nl
*   \_______    /                       \  ___/mail      Bitnet: rcrolf@heitue5
* _________ \  / _______________________ \_____> ______________________________
*    The RvG \/ Clipper Collection Home Page: http://asterix.urc.tue.nl/~rcrolf
*
*******************************************************************************

*: EOF: RPROFILE.PRG
